# Copyright (c) Ran Dugal 2014
#
# This file is part of dust.
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

''' invoke commands or a shell over ssh sessions,  demultiplex the ssh output '''  

import getpass
from threading import Thread
import select
import socket
import sys
import os, struct, fcntl
import colorama

from paramiko.py3compat import u
import paramiko

from dustcluster.util import setup_logger
logger = setup_logger( __name__ )


# Once a session has been setup a program at the remote end can be  
# executed with SSH_MSG_CHANNEL_REQUEST, with string 'shell', 'exec', or 
# 'subsystem' for a default shell, single command, or subsystem. 
# Here we start a full interactive shell and invoke commands on it, 
# demultiplexing the output from all chanells onto stdout locally. 
# This allows us to execute arbitrarily interactive scripts. 

class ReceiveDemux(object):
    ''' receive demultiplexer for all open ssh interactive shells ''' 

    def __init__(self, session_mgr):
        self.refresh_callback = None
        self.chans = {} # { chan : receive_buffer }
        self.session_mgr = session_mgr
        self.state = 'created'
        self.thread = Thread(target=self.receive_loop)
        self.thread.daemon = True
        self.thread.start()

    def start(self, sshterm):
        ''' start demuxing output on this term '''
        self.chans[sshterm.chan] = sshterm

    def stop(self, chan):
        ''' stop receiving on chan, remove session ''' 
        term = self.chans[chan]
        del self.chans[chan]
        self.session_mgr.remove_session(term)

    def shutdown(self):
        ''' shut down receiver thread '''
        self.state = 'shutdown'
        self.thread.join()

    def handle_read(self, achan):
        try:
            readbytes = u(achan.recv(1024))
            if len(readbytes) == 0:
                sys.stdout.write('\r\SSH session disconnected.\r\n')
                sys.stdout.flush()
                self.stop(achan)
                return

            sshterm = self.chans[achan]
            if sshterm.raw_shell_mode:
                sys.stdout.write(readbytes)
                sys.stdout.flush()
                
            else:
                sshterm.recvbuf = sshterm.recvbuf + readbytes

        except socket.timeout:
            sys.stdout.write('\r\SSH session timedout.\r\n')
            sys.stdout.flush()
            self.stop(achan)

    def handle_err(self, achan):
        try:
            self.stop(achan)
        except:
            logger.exception('Error on socket, could not shutdown cleanly.\r\n')


    def handle_timeout(self):
    
        try:

            wrote_output = False
            for chan, sshterm in  self.chans.iteritems():

                if not sshterm.recvbuf:
                    continue

                if sshterm.raw_shell_mode:
                    continue

                if not sshterm.login_guid_found:
                # surpress login banner. disabled for now.
                # Note: RFC-4254 reccomends the use of magic cookeis to surpress spurious
                # output when starting a subsystem via the shell "to distinguish it from 
                # arbitrary output generated by shell initialization scripts, etc. This spurious 
                # output from the shell may be filtered out either at the server or at the client"
                # We can reuse this trick to surpress login banner text and echo enable/disable.
                    pos1 = sshterm.recvbuf.find( '\n' + SSHTerm.login_complete_guid )
                    if pos1 != -1:
                        sshterm.login_guid_found = True
                        guid_len = len(SSHTerm.login_complete_guid)  + 1
                        sshterm.recvbuf = sshterm.recvbuf[pos1 + guid_len:]
                    else:
                        continue

                if sshterm.recvbuf.strip():
                    sys.stdout.write('\n')
                    prefix = "\n%s%s[%s]%s " % (colorama.Style.BRIGHT, colorama.Fore.WHITE, sshterm.node.name, 
                                                colorama.Style.RESET_ALL)
                    sys.stdout.write(prefix)
                    sys.stdout.write(sshterm.recvbuf.replace('\n', prefix))
                    sys.stdout.flush()
                    sshterm.recvbuf = u('')
                    sys.stdout.write('\n')
                    sys.stdout.flush()
                    wrote_output = True

            #reset prompt
            if wrote_output and self.refresh_callback:
                self.refresh_callback()

        except: 
            logger.exception('Error on receive loop, ssh session in bad state.\r\n')


    def receive_loop(self):
        ''' demux receive loop '''


        while self.state != 'shutdown':
            r, w, e = select.select(self.chans, [], self.chans, 0.25)
            if r:
                for achan in r:
                    self.handle_read(achan)
            if e:
                for achan in e:
                    self.handle_err(achan)

            if not r and not e:
                self.handle_timeout()

        logger.debug('Exiting receive loop.\r\n')

        return 0


class SessionManager(object):
    ''' holds a map of node ids to ssh sessions
        registers/unregisters ssh sessions with the demultiplexer 
    '''

    def __init__(self):
        self.demux = ReceiveDemux(self)
        self.session_map = {}

    def remove_session(self, term):
    
        if term.raw_shell_mode:
            logger.info('Logged out of ssh session. Press enter to continue.\n\r')
            term.raw_shell_mode = True
            term.revert_tty()

        for nodeid, nodeterm in self.session_map.items():
            if nodeterm == term:
                del self.session_map[nodeid]

    def shutdown(self):
        for term in self.session_map.values():
            term.shutdown()

        self.demux.shutdown()

    def term_from_node(self, node, keyfile, rawshell=False):

        term = self.session_map.get(node.get('id'))

        if not term:
            term = SSHTerm(node, keyfile)
            # hide login banner unless this is a raw shell login
            cookie = False
            if not rawshell:
                term.login_guid_found = False
                cookie = True
            term.login(cookie)
            self.demux.start(term)
            self.session_map[node.get('id')] = term

        if not term.is_connected():
            logger.info('no ssh connection, logging in')
            term.login()

        return term


class SSHTerm(object):
    '''
    Ssh client session - starts an interactive ssh shell
    Takes line input commands or enters a char bufferred raw terminal
    '''

    prompt = "dust:ssh:%s:$ "

    login_complete_guid = 'B79D8677-F58A-4E09-B917-855A6619A951' # GUID

    def __init__(self, node, keyfile):
        self.prompt = "dust:ssh:%s:$ " % node.name
        self.node = node
        self.keyfile = keyfile

        self.state = 'not_connected'
        self.transport  = None
        self.chan = None
        self.newchan = None

        self.recvbuf = u('')
        self.login_guid_found = True

        self.raw_shell_mode = False
        self.oldattrs  = None

        self.sftp = None # sftp subservice

        self.echo  = True

    def is_connected(self):
        return self.transport and self.transport.is_authenticated() and self.transport.is_active()

    def login(self, cookie=False):
        hostname = self.node.get('public_ip_address') or self.node.get('private_ip_address')
        if not hostname:
            hostname = self.node.get('hostname')
        username = self.node.login_rule.get('login-user')

        if not hostname or not username:
            raise Exception("No ip, hostname or username available for node %s" % node.name or node.index)

        logger.debug('hostname=[%s], username=[%s], key=[%s]' % (hostname, username, self.keyfile))
        if not self.is_connected():
            try:
                self.connect(hostname, username)
                self.transport.set_keepalive(60*3)
                self.state = 'connected'
                if cookie:
                    self.disable_echo(auxcmd= "; echo %s" % self.login_complete_guid)
                    self.echo = False
            except:
                logger.error('error on ssh login on host %s :' % (hostname))
                raise

    def disable_echo(self, auxcmd=''):
        logger.debug( '%s: disabling echo' % self.node.name )
        cmd = "stty -echo; export PS1=''"
        self.command(cmd + auxcmd)
        self.echo = False

    def enable_echo(self, auxcmd=''):
        logger.info( '%s: enabling echo' % self.node.name )
        #prompt = "\[\033[01;32m\]\u@\h\[\033[00m\]:\[\033[01;34m\]\w\[\033[00m\]\$ "
        prompt = "rawshell:%s:\w\$ " % self.node.name
        cmd = "stty echo; export PS1='%s'" % prompt
        self.command(cmd + auxcmd)
        self.echo = True

    def raw_shell_other(self):
        print "raw mode shell not suppported on this system yet."
        return

    def raw_shell(self):

        if os.name == 'posix':
            self.raw_shell_posix()
        else:
            self.raw_shell_other()

    def raw_shell_posix(self):

        import termios
        import tty

        self.raw_shell_mode = True

        self.oldattrs = termios.tcgetattr(sys.stdin)
        tty.setraw(sys.stdin.fileno())
        tty.setcbreak(sys.stdin.fileno())
        self.chan.settimeout(0.0)

        # set pty size
        s = struct.pack ("HHHH", 0, 0, 0, 0)
        sz = struct.unpack ('HHHH', fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ , s))
        self.chan.resize_pty(sz[1], sz[0])

        ctrlc_count = 0

        ctrlout = False
        while self.raw_shell_mode:
            try:
                d = sys.stdin.read(1)
                if not d:
                    break

                if ord(d) == 3:
                    ctrlc_count += 1
                else:
                    ctrlc_count = 0

                if ctrlc_count > 2:
                    ctrlout = True
                    break

                ret = self.chan.send(d)

                if (not ret):
                    logger.error('ssh session closed.')
                    break
            except socket.error:
                logger.info("socket closed")
            except:
                logger.exception('exception in raw shell:')

        if ctrlout:
            logger.debug( '%s: switching back to line buffered commands' % self.node.name )
            self.disable_echo()

        self.revert_tty()

    def revert_tty(self):
        ''' revert tty attribute '''
        import termios
        self.raw_shell_mode = False
        if ( self.oldattrs ):
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.oldattrs)

    def shutdown(self):
        ''' shutdown this ssh term '''
        self.state = 'shutdown'
        if self.chan:
            self.chan.close()
        if self.transport:
            self.transport.close()
        logger.info( '%s: closed ssh' % self.node.name )

    def command(self, line):
        ''' send a shell command to the interactive ssh shell '''

        if self.state != 'connected':
            logger.info( 'session not connected' )
            return

        if not self.is_connected():
            logger.info( 'ssh session not connected, authed, or active' )
            return

        self.chan.send(line)
        self.chan.send('\n')

    #TODO: override port from template
    def connect(self, hostname, username, port=22):
        ''' connect and authenticate ''' 

        private_key_path = self.keyfile 

        logger.info('ssh login to host=[%s] keyfile=[%s] user=[%s]' % (hostname, private_key_path, username))

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((hostname, port))

        self.transport = paramiko.Transport(sock)
        self.transport.start_client()

        #TODO: check host key

        try:
            key = paramiko.RSAKey.from_private_key_file(private_key_path)
        except paramiko.PasswordRequiredException:
            password = getpass.getpass('RSA key password: ')
            key = paramiko.RSAKey.from_private_key_file(private_key_path, password)

        if key:
            self.transport.auth_publickey(username, key)

        if not self.transport.is_authenticated():
            self.transport.close()
            raise Exception('Authentication failed.')

        self.chan = self.transport.open_session()
        self.chan.get_pty()
        self.chan.invoke_shell()



class LineTerm(object):
    '''
    top level api - implements ssh and raw terminal functionality for a set of nodes 
    '''

    def __init__(self):
        self.session_manager = SessionManager()

    def set_refresh_callback(self, callback):
        '''Optional callback after a block of ssh output is written to stdout. 
            for commands issued in interactive mode this need not be the end of output 
        '''
        self.session_manager.demux.refresh_callback = callback

    def command(self, keyfile, node, cmd=None):
        ''' send a command to an interactive ssh shell or enter a raw shell input loop.
            in both cases log in if not logged in. 
        '''

        term = None
        try:
            rawshell= False if cmd else True
            term = self.session_manager.term_from_node(node, keyfile, rawshell=rawshell)

            if cmd:
                if term.echo:
                    term.disable_echo()
                term.command(cmd)
            else:
                if not term.echo:
                    term.enable_echo()
                term.raw_shell()
        except Exception, e:
            logger.error('Dust: Error: %s' % e)
            raise
        finally:
            if term:
                term.revert_tty()

    def shell(self, keyfile, node):

        logger.info(\
         '*** Entering raw shell, press ctrl-c thrice to return to cluster shell. Press Enter to continue.***')

        raw_input()

        self.command(keyfile, node, cmd=None)

    def put(self, keyfile, node, srcfile, destdir=None):

        if not os.path.isfile(srcfile):
            logger.error('file does not exist locally : %s' % srcfile)
            return

        try:
            term = self.session_manager.term_from_node(node, keyfile)
            if not term.sftp:
                term.sftp = paramiko.SFTPClient.from_transport(term.transport)

            sftp = term.sftp
            fname = os.path.basename(srcfile)
            if not destdir:
                destfile = fname
            else:
                destfile = destdir + "/" + fname
            ret = sftp.put(srcfile, destfile, confirm=True)

            if not getattr(ret, 'filename', None):
                ret.filename = os.path.basename(destfile)
            
            logger.info('uploaded to %s : %s' % (node.name, ret))
        except Exception, e:
            logger.error(e)


    def get(self, keyfile, node, remotefile, localdir):

        if localdir and not os.path.isdir(localdir):
            logger.error('dir does not exist locally : %s' % localdir)
            return

        try:
            term = self.session_manager.term_from_node(node, keyfile)
            if not term.sftp:
                term.sftp = paramiko.SFTPClient.from_transport(term.transport)

            fname = os.path.basename(remotefile)
            if localdir:
                localfile = os.path.join(localdir, fname)
            else:
                localfile = fname

            localfile = '%s.%s' % (localfile, node.name)

            sftp = term.sftp
            logger.info('getting %s' % (remotefile))

            sftp.get(remotefile, localfile)

            logger.info('downloaded from %s : %s' % (node.name, localfile))
        except Exception, e:
            logger.error(e)

    def shutdown(self):
        self.session_manager.shutdown()





# Copyright (c) Ran Dugal 2014
#
# This file is part of dust cluster
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import logging
import colorama

''' dummy command to handle $help filters'''

help_str = '''
Most commands take a filter. e.g. tag filter key=val
Here, "filter" is a comma separated list of filter expressions.

Filter expression is evaluated as (in order of precedence):

**A node index from "show"**
    dust$ stop 1,2,3
    dust$ @1,2,3 service restart nginx

**A cluster name**
    dust$ show -v pyspark1      # all nodes in cluster pyspark1
    dust$ stop slurm1           # all nodes in cluster slurm1
    dust$ @slurm1 uptime        

    Note: see $help assign for assigning nodes to clusters

**A node name**
    Name comes from Instance tags with Key=Name.e.g { Name : mysql }
    dust$ stop worker*         # names can have wildcards
    dust$ start wo*
    dust$ terminate worker[0-2]
    dust$ @worker* vmstat

**attr=value EC2 attribute or tags**
    dust$ start state=stopped
    dust$ start state=stop*       # filters can have wildcards
    dust$ @ip=54.12.* uptime
    dust$ @tags=owner:devops uptime
    dust$ stop tags=env:*       # tags can have wildcards too
    dust$ show -v launch_time=2016-04-03*

Note: in addition, the show command takes a search term so  
        $show -v 192.168
      does a free text search for "192.168" in all attributes/tags.
'''

# export commands 
commands  = ['filters']

def filters(cmdline, cluster, logger):
    '''
    filters - see help on filter expressions
    '''

    print "%s%s%s" % (colorama.Fore.GREEN, help_str, colorama.Style.RESET_ALL)




import pprint

commands = ['tag', 'untag', 'openport']

def tag(cmdline, cluster, logger):
    '''
    tag filter key=value   - add tag to target nodes

    Notes:
    tag an ec2 node with key=value

    Examples:
    tag worker* env=dev    # add tag env=dev to nodes named worker* 

    Multiple tags works.
    tag worker* tag1=val1,tag2=val2,tag3=val3 - add tags tag1,tag2,tag3 on target nodes
    '''

    try:

        args = cmdline.split()

        if len(args) < 2:
            logger.error("usage: tag target tag=value")
            return

        target = args[0]
        tags   = args[1]

        taglist = tags.split(",")

        target_nodes = cluster.any_nodes_from_target(target)
        if not target_nodes:
            logger.info('No running nodes found.')
            return

        client = cluster.cloud.client()
        r_ids = [node.get('id') for node in target_nodes]

        tagparam = get_tag_param(taglist)

        client.create_tags(Resources=r_ids, Tags=tagparam)

        # refresh from cloud next operation
        cluster.invalidate_cache()

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

    logger.info( 'ok' )




def untag(cmdline, cluster, logger):
    '''
    untag filter key=value   - remove a tag from a target node

    Notes:
    untag an ec2 node with tag key and value 

    Examples:
    untag worker* env=dev           # remove tag env from nodes named worker*
    untag state=running env=prod
    '''

    try:

        args = cmdline.split()

        if len(args) < 2:
            logger.error("usage: untag filter-exp tag=value")
            return

        target = args[0]
        tags   = args[1]

        taglist = tags.split(",")

        target_nodes = cluster.any_nodes_from_target(target)
        if not target_nodes:
            logger.info('No running nodes found.')
            return

        client = cluster.cloud.client()
        r_ids = [node.get('id') for node in target_nodes]

        tagparam = get_tag_param(taglist)

        client.delete_tags(Resources=r_ids, Tags=tagparam)

        # refresh from cloud next operation
        cluster.invalidate_cache()

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

    logger.info( 'ok' )

def get_tag_param(taglist):

        tagparam = []

        for tag in taglist:

            if "=" in tag:
                tagkey, tagval = tag.split("=")
            else:
                tagkey = tag
                tagval = ""

            if tagkey.startswith("aws"):
                logger.error("Error: tagname cannot start with [aws].")
                return

            tagparam.append( {'Key': tagkey, 'Value': tagval } )

        return tagparam


def openport(cmdline, cluster, logger):
    '''
    openport secgroup_id [port  cidr_range]   - list or open TCP ingress port

    secgroup_ip: security group id. (see show -vv under security_groups)
    port: ingress port to open
    cidr_range: cidr range or 0.0.0.0/0 (all) if not given

    Examples:                   
    openport sg-e12345                         # list ingress rules        
    openport sg-e12345 8080                    # open ingress tcp port 8080 from all ips
    openport sg-e12345 8080 192.168.3.0/24     # open ingress tcp port 8080 from cidr range 
    '''

    try:

        args = cmdline.split()

        if len(args) < 1:
            logger.error("openport sec_grp [port] [cidr range]")
            return

        grp_id  = args[0].strip()
        res = cluster.cloud.conn()
        grp = res.SecurityGroup(grp_id)

        if len(args) < 2:
            show_ingress_rules(grp)
            return

        port    = args[1].strip()
        range   = args[2].strip() if len(args) > 2 else "0.0.0.0/0"

        grp.authorize_ingress(CidrIp=range, FromPort=int(port), ToPort=int(port), IpProtocol='tcp')

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

    logger.info( 'ok' )

def show_ingress_rules(grp):
        grp.reload()
        for perm in grp.ip_permissions:
            pprint.pprint(perm)





'''
commands to view cloudwatch logs easily
'''
import getopt
import time
import boto3
import colorama

commands = ['logs']

client_cache = {}

# if the arg matches a known log group/stream dont search for it
known_log_groups = set()
known_log_streams = set()

def get_client(cluster):
    client = client_cache.get(cluster.cloud.region)
    if not client:
        creds = cluster.config.get_credentials()
        client = boto3.client('logs', region_name=cluster.cloud.region, 
                                            aws_access_key_id=creds.get('aws_access_key_id'),
                                            aws_secret_access_key=creds.get('aws_secret_access_key'))

    return client

def logs(cmdline, cluster, logger):
    '''
    logs [-N M] [-t] [log_group] [log_stream]  - show M last cloudwatch log groups, log streams, or log events

    Notes:
    -N M  show last M entries
    -t    tail the most recent events from a log group

    Examples:
    logs                                show all log groups
    logs /aws/lam                       show log groups starting with /aws/lam
    logs /aws/lambda/test1              show log streams in this log group
    logs -t /aws/lambda/test1           show the most recent event streams events
    logs -t -N  50  /aws/lambda/f1      show the last 50 most recent events in this log group 
    logs -N 10 log_group log_stream     show last 10 events in log_stream
    '''
    opts,args = getopt.getopt(cmdline.split(), "tN:")

    limit = 5
    do_tail = False
    for opt, optval in opts:
        if opt == '-N':
            limit = int(optval)
        elif opt == '-t':
            do_tail = True

    loggroup_name = ""
    logstream_name = ""
    if len(args) > 0:
        loggroup_name = args[0]
    if len(args) > 1:
        logstream_name = args[1]

    resolved_log_group = ""

    client = get_client(cluster)

    print colorama.Fore.GREEN, "limit=%s" % limit, colorama.Style.RESET_ALL

    if loggroup_name in known_log_groups:
        logger.debug("known log group")
        resolved_log_group = loggroup_name
    else:
        logger.info("searching ...")
        # show log groups
        log_groups = None
        if loggroup_name:
            log_groups = client.describe_log_groups(logGroupNamePrefix=loggroup_name, limit=limit)
        else:
            log_groups = client.describe_log_groups(limit=limit)

        print colorama.Fore.GREEN, "log groups:", colorama.Style.RESET_ALL
        log_groups = log_groups.get('logGroups')
        log_groups = sorted(log_groups, key= lambda x: x.get('creationTime'))
        for log_group in log_groups:
            lg_name = log_group.get('logGroupName')
            ts = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(long(log_group['creationTime'])/1000.0))
            print ts, colorama.Fore.CYAN, lg_name, colorama.Style.RESET_ALL
            known_log_groups.add(lg_name)
            if len(log_groups) == 1 and lg_name == loggroup_name:
                resolved_log_group = loggroup_name

    if not resolved_log_group:
        logger.debug("could not resolve a unique log group. returning")
        return

    resolved_log_stream = ""
    if logstream_name in known_log_streams:
        logger.debug("known log stream")
        resolved_log_stream = logstream_name
        log_streams = []
    else:
        if logstream_name:
            log_streams = client.describe_log_streams(logGroupName=resolved_log_group, 
                    logStreamNamePrefix=logstream_name, limit=limit)
        else:
            log_streams = client.describe_log_streams(logGroupName=resolved_log_group, 
                    orderBy='LastEventTime', descending=True, limit=limit if not do_tail else 3)

        print colorama.Fore.GREEN, "log streams:", colorama.Style.RESET_ALL
        log_streams = log_streams.get('logStreams')

        print "%-20s %-20s Messsage" % ('Creation', 'Last Event')
        for log_stream in log_streams:
            ls_name = log_stream.get('logStreamName')
            ts1 = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(long(log_stream['creationTime'])/1000.0))
            ts2 = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(long(log_stream.get('lastEventTimestamp'))/1000.0))
            print ts1, ts2, colorama.Fore.CYAN, ls_name, colorama.Style.RESET_ALL

            known_log_streams.add(ls_name)
            if len(log_streams) == 1 and ls_name == logstream_name:
                resolved_log_stream = ls_name

        if do_tail:
            resolved_log_stream = log_streams[-1].get('logStreamName')

    if do_tail and logstream_name:
        for log_stream in log_streams:
                show_log_events(client, resolved_log_group, log_stream.get('logStreamName'), 3)
        return

    if not resolved_log_stream:
        logger.debug("could not resolve a unique log stream. returning")
    else:
        show_log_events(client, resolved_log_group, resolved_log_stream, limit)



def show_log_events(client, resolved_log_group, resolved_log_stream, limit):
    print colorama.Fore.GREEN, "log events from %s:" % resolved_log_stream, colorama.Style.RESET_ALL

    log_events = client.get_log_events(logGroupName = resolved_log_group,
                                        logStreamName = resolved_log_stream, limit=limit)

    for log_event in log_events.get('events'):
        ts = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(long(log_event['timestamp'])/1000.0))
        msg = log_event['message']
        if msg[-1] == '\n':
            msg = msg[:-1]
        if msg[0] in ['{', '[']:
            msg = format_as_json(msg)
        print ts, colorama.Fore.CYAN, msg, colorama.Style.RESET_ALL
    

def format_as_json(msg):
    ''' assume invalid json, pretty print dont decode '''
    ret = ""
    indent = -1
    for c in msg:
        ret += c
        if c == ',':
            ret += '\n'
            ret += '    ' * indent
        if c in ['{', '[']:
            indent += 1
            ret += '\n'
            ret += '    ' * indent
        if c in ['}', ']']:
            indent -= 1 
    return ret





import os
import colorama
import ConfigParser
from EC2 import EC2Config
import stat
import yaml
import glob

from dustcluster.util import setup_logger
logger = setup_logger( __name__ )

class DustConfig(object):
    ''' loads, saves, owns credentials config file and user_data config file '''

    _instance = None
    _inited   = False

    def __new__(cls):
        if not DustConfig._instance:
             DustConfig._instance = object.__new__(cls)
        return DustConfig._instance

    user_dir         = os.path.expanduser('~')
    dust_dir         = os.path.join(user_dir, '.dustcluster')

    def __init__(self):

        if DustConfig._inited:
            return

        DustConfig._inited = True

        self.credentials = {}
        self.user_data   = {}
        self.login_rules = []
        self.clusters    = {}

        self.history_file     = os.path.join(self.dust_dir, 'cmd_history')
        self.credentials_file = os.path.join(self.dust_dir, 'credentials')
        self.userdata_file    = os.path.join(self.dust_dir, 'user_data')
        self.aws_credentials_file  = os.path.join(self.user_dir, '.aws/credentials')
        self.login_rules_file = os.path.join(self.dust_dir, 'login_rules.yaml')
        self.default_keys_dir = os.path.join(self.dust_dir, 'keys')
        self.clusters_dir     = os.path.join(self.dust_dir, 'clusters')

        if not os.path.exists(self.dust_dir):
            os.makedirs(self.dust_dir)

        # first time setup
        if not os.path.exists(self.credentials_file) or not os.path.exists(self.userdata_file):
            logger.info("%sWelcome to dustcluster, creating config file:%s"  % (colorama.Fore.GREEN, colorama.Style.RESET_ALL))

            aws_creds = {}
            if os.path.exists(self.aws_credentials_file):
                confirm = raw_input("Found default awscli credentials in [%s]. Use these? [Y]:" % self.aws_credentials_file) or "y"
                if confirm[0].lower() == "y":
                    aws_creds = self.read_credentials(self.aws_credentials_file)

            ec2creds, ec2data =  EC2Config.setup_credentials(aws_creds)
            self.credentials.update(ec2creds)
            self.user_data.update(ec2data)
            self.write_credentials()
            self.write_user_data()
        else:
            # read in credentials and user data
            logger.debug("Reading credentials from [%s]" % self.credentials_file)
            self.credentials = self.read_credentials(self.credentials_file)

            logger.debug("Reading user data from [%s]" % self.userdata_file)
            self.user_data = self.read_user_data()

            EC2Config.validate(self.credentials, self.user_data)

            self.read_all_clusters()
            self.read_login_rules()

    @staticmethod
    def get():
        return DustConfig._instance

    def get_credentials(self):
        return self.credentials

    def get_default_keys_dir(self):
        return self.default_keys_dir

    def get_userdata(self):
        return self.user_data

    def get_login_rules(self):
        return self.login_rules

    def get_clusters(self):
        return self.clusters

    def get_history_file_path(self):
        return self.history_file

    def get_clusters_dir(self):

        if not os.path.exists(self.clusters_dir):
            os.makedirs(self.clusters_dir)

        return self.clusters_dir

    def read_login_rules(self):
        if os.path.exists(self.login_rules_file):
            with open(self.login_rules_file, "r") as fh:
                rules = yaml.load(fh.read())
                login_rules = rules
        else:
            login_rules = {}

        self.login_rules =  login_rules.get('login_rules') or []

    def write_login_rules(self):
        str_yaml = yaml.dump({ "login_rules" : self.login_rules}, default_flow_style=False)
        with open(self.login_rules_file, 'w') as fh:
            fh.write(str_yaml)

    def read_user_data(self):
        if os.path.exists(self.userdata_file):
            with open(self.userdata_file, 'r') as fh:
                user_data = yaml.load(fh) or {}
        else:
            user_data = {}

        return user_data

    def write_user_data(self):

        logger.debug("Writing user data to %s" % self.userdata_file)

        str_yaml = yaml.dump(self.user_data, default_flow_style=False)

        if not os.path.exists(self.dust_dir):
            os.makedirs(self.dust_dir)

        with open(self.userdata_file, 'w') as yaml_file:
            yaml_file.write(str_yaml)

    def read_credentials(self, file_path):
        ''' read dustcluster or aws credentials '''

        parser = ConfigParser.ConfigParser()
        parser.read(file_path)

        config_data = parser.defaults()

        if not config_data:
            config_data = dict(parser.items("default"))

        return config_data

    def write_credentials(self):

        # write creds
        logger.info("Writing credentials to [%s] with mode (0600)" % self.credentials_file)
        parser = ConfigParser.ConfigParser(self.credentials)
        if not os.path.exists(self.dust_dir):
            os.makedirs(self.dust_dir)

        with open(self.credentials_file, 'wb') as fh:
            parser.write(fh) 
        os.chmod(self.credentials_file, stat.S_IRUSR | stat.S_IWUSR)

    def save_cluster_config(self, name, str_yaml):
        ''' return path if saved ''' 

        template_file = "%s.yaml" % name
        template_file = os.path.join(self.clusters_dir, template_file)

        if os.path.exists(template_file):
            yesno = raw_input("%s exists. Overwrite?[y]:" % template_file) or "yes"
            if not yesno.lower().startswith("y"):
                return None

        if not os.path.exists(self.clusters_dir):
            os.makedirs(self.clusters_dir)

        str_comment = "#This picks up the latest Amazon Linux AMI with user ec2-user.\n" + \
                      "#For a custom AMI, add props under cluster: \n" + \
                      "#   image: amixyz\n" + \
                      "#   username: login user\n\n"
        with open(template_file, 'w') as yaml_file:
            yaml_file.write(str_comment)
            yaml_file.write(str_yaml)

        return template_file

    def delete_cluster_config(self, name, region):
        template_file = "%s.yaml" % name
        template_file = os.path.join(self.clusters_dir, template_file)
        os.remove(template_file)

        template_file = "%s.%s.cfn" % (region,name)
        template_file = os.path.join(self.clusters_dir, template_file)
        if os.path.exists(template_file):
            os.remove(template_file)

        logger.info("Deleted cluster config: %s" % template_file)

    def read_all_clusters(self):
        wildcardpath = os.path.join(self.clusters_dir, "*.yaml")
        cluster_files = glob.glob(wildcardpath)
        logger.debug("found [%d] clusters in %s" % (len(cluster_files), wildcardpath))
        clusters = {}
        for cluster_file in cluster_files:
            
            if os.path.isfile(cluster_file):
                with open(cluster_file, "r") as fh:
                    cluster = yaml.load(fh.read())
                    cluster_props = cluster.get('cluster')
                    cluster_name = cluster_props.get('name')
                    clusters[cluster_name] = cluster

        self.clusters = clusters




import time
import datetime
import boto3

commands = ['rules']
client_cache = {}

def get_client(cluster):
    client = client_cache.get(cluster.cloud.region)
    if not client:
        creds = cluster.config.get_credentials()
        client = boto3.client('config', region_name=cluster.cloud.region, 
                                            aws_access_key_id=creds.get('aws_access_key_id'),
                                            aws_secret_access_key=creds.get('aws_secret_access_key'))
    return client


def run_rule(rulename, cluster, logger):

    client = get_client(cluster)
    client.start_config_rules_evaluation(ConfigRuleNames=[rulename])

    for i in range(12):
        time.sleep(1)
        print "."
    print

    # get custom lambda name
    ret = client.describe_config_rules(ConfigRuleNames=[rulename])
    source = ret['ConfigRules'][0].get('Source')
    lambda_name = ""
    if source.get('Owner') == 'CUSTOM_LAMBDA':
        lambda_name = source.get('SourceIdentifier').split(":")[-1]
    
    if lambda_name:
        log_group = '/aws/lambda/' + lambda_name
        logger.info("logs -t " + log_group)
        cluster.handle_command("logs", "-t " + log_group)
    else:
        logger.info("No custom lambda found. %s" % source)


def get_max_date(lastgood, lastfailed):
    if lastgood and lastfailed:
        last_invoked = max(lastgood, lastfailed)
    else:
        last_invoked = lastgood or lastfailed

    return last_invoked


def rules_status(rulename, cluster, logger):

    client = get_client(cluster)

    if rulename:
        resp = client.describe_config_rule_evaluation_status(ConfigRuleNames=[rulename], 
                        Limit=50)
    else:
        resp = client.describe_config_rule_evaluation_status(Limit=50)

    fmt = "%-25s %-30s %-30s %-10s %-10s"
    print (fmt % ("Rule name", "Last Invoked", "Last Eval", "Last Error", "Last Err Msg"))

    status_list = resp.get('ConfigRulesEvaluationStatus')
    status_list = filter(
                         lambda x :  get_max_date(x.get('LastSuccessfulInvocationTime'),x.get('LastFailedInvocationTime')),
                         status_list
                         )

    status_list = sorted(status_list, 
                        key= lambda x : get_max_date(x.get('LastSuccessfulInvocationTime'),
                        x.get('LastFailedInvocationTime')))

    for status in status_list:
        last_invoked = get_max_date(status.get('LastSuccessfulInvocationTime'),
                                    status.get('LastFailedInvocationTime'))

        last_eval = get_max_date(status.get('LastSuccessfulEvaluationTime'), 
                                 status.get('LastFailedEvaluationTime'))

        print (fmt % (status['ConfigRuleName'], last_invoked, last_eval,
                                             status.get('LastErrorCode'), status.get('LastErrorMessage')))

def rules(cmdline, cluster, logger):
    '''
    rules [run rulename | status rulename]    - run config rule rule_name, and print logs

    Examples:
    rules run hello_world
    rules status hello_world
    '''

    args = cmdline.split()
    if len(args) < 1:
        print "usage: rules [run | status] rule name"
        return

    cmd = args[0]
    rulename = ""
    if len(args) > 1:
        rulename = args[1]

    if (cmd == "run"):
        if not rulename:
            print "no rulename"
            return
        logger.info("running rule %s" % rulename)
        run_rule(rulename, cluster, logger)
    elif (cmd == "status"):
        rules_status(rulename, cluster, logger)




import pprint
from boto import cloudformation
from troposphere import ec2, Ref, Template, GetAtt
import yaml
import boto
import copy
import os
import colorama
from dustcluster.EC2 import EC2Config

commands = ['cluster']

def usage(logger):
    logger.error("cluster [new] | [create clustername] | [status clustername] | [delete clustername] | list")


def cluster(cmdline, cluster, logger):
    '''
    cluster list|new|create|status|delete - list, create, delete or print cluster status

    Commands:

    cluster new
    cluster create clustername
    cluster status clustername
    cluster delete clustername
    cluster list

    Examples:

    cluster new                 # create a new cluster spec
    cluster create              # bring up a cluster using cloudformation 
    cluster status slurm
    cluster delete slurm
    cluster list                # see all cloudformation clusters
    '''

    args = cmdline.split()

    if not args:
        usage(logger)
        return

    subcommand = args[0]

    if subcommand == 'new':
        new_cluster(args, cluster, logger)
    elif subcommand == 'create':
        create_cluster(args, cluster, logger)
    elif subcommand == 'delete':
        delete_cluster(args, cluster, logger)
    elif subcommand == 'status':
        cluster_status(args, cluster, logger)
    elif subcommand == 'list':
        list_clusters(args, cluster, logger)
    else:
        usage(logger)
        return

def list_clusters(args, cluster, logger):
    cluster.show_clusters()

def get_closest_region(cluster, logger):

    user_data = cluster.config.get_userdata()
    region = user_data.get('closest_region')
    if region:
        return region

    creds = cluster.config.get_credentials()
    region = EC2Config.find_closest_region(logger, creds.get('aws_access_key_id'), creds.get('aws_secret_access_key'))
    user_data['closest_region'] = region
    cluster.config.write_userdata()

    return region

def configure_security_groups(vpc_subnet, vpc_id):

    node_sec_group = ec2.SecurityGroup('DustNodeSG')
    node_sec_group.GroupDescription = "Allow incoming ssh and icmp and all intracluster tcp"
    if vpc_subnet:
        node_sec_group.VpcId = vpc_id

    ssh_in_rule = ec2.SecurityGroupRule()
    ssh_in_rule.FromPort = 22
    ssh_in_rule.ToPort = 22
    ssh_in_rule.CidrIp = "0.0.0.0/0"
    ssh_in_rule.IpProtocol = 'tcp'

    icmp_in_rule = ec2.SecurityGroupRule()
    icmp_in_rule.FromPort = -1
    icmp_in_rule.ToPort = -1
    icmp_in_rule.CidrIp = "0.0.0.0/0"
    icmp_in_rule.IpProtocol = 'icmp'

    node_sec_group.SecurityGroupIngress = [ssh_in_rule, icmp_in_rule]

    intracluster_sec_group = ec2.SecurityGroupIngress('DustIntraClusterSG')
    if vpc_subnet:
        intracluster_sec_group.GroupId = Ref(node_sec_group)
    else:
        intracluster_sec_group.GroupName = Ref(node_sec_group)
    intracluster_sec_group.FromPort = 0 
    intracluster_sec_group.ToPort = 65535
    if vpc_subnet:
        intracluster_sec_group.SourceSecurityGroupId = Ref(node_sec_group)
    else:
        intracluster_sec_group.SourceSecurityGroupName = Ref(node_sec_group)
    intracluster_sec_group.IpProtocol = "-1"

    return node_sec_group, intracluster_sec_group 

def configure_vpc(cfn_template, cluster_name):

    vpc = ec2.VPC("DustVPC")
    vpc.CidrBlock = "10.0.0.0/16"
    vpc.Tags = [ec2.Tag("Name:", cluster_name)]
    cfn_template.add_resource(vpc)
    vpc_id = Ref(vpc)

    subnet = ec2.Subnet('dustSubnet')
    subnet.VpcId = vpc_id
    subnet.CidrBlock = "10.0.0.0/24"
    cfn_template.add_resource(subnet)
    vpc_subnet = Ref(subnet)

    net_gateway = ec2.InternetGateway('dustGateway')
    cfn_template.add_resource(net_gateway)

    attach_net_gateway = ec2.VPCGatewayAttachment('dustAttachGateway')
    attach_net_gateway.VpcId = vpc_id
    attach_net_gateway.InternetGatewayId = Ref(net_gateway)
    cfn_template.add_resource(attach_net_gateway)

    route_table = ec2.RouteTable('dustRoutetable')
    route_table.VpcId = vpc_id
    cfn_template.add_resource(route_table)

    route = ec2.Route('dustRoute')
    route.RouteTableId = Ref(route_table)
    route.DestinationCidrBlock = "0.0.0.0/0"
    route.GatewayId = Ref(net_gateway)
    route.DependsOn = "dustAttachGateway"
    cfn_template.add_resource(route)

    attach_route = ec2.SubnetRouteTableAssociation('dustAttachRouteTable')
    attach_route.SubnetId = vpc_subnet
    attach_route.RouteTableId = Ref(route_table)
    cfn_template.add_resource(attach_route)

    return vpc_id, vpc_subnet

def new_cluster(args, cluster, logger):

    name = raw_input("Name this cluster: ")

    if (os.path.exists(cluster.config.get_clusters_dir() + "%s.cfn" % name)):
        clusters = cluster.config.get_clusters()
        logger.error("A cluster named %s already exists in region %s. Please use a different name" % 
                                    (name, clusters.get(name).get('cloud').get('region')))
        return

    numnodes = int(raw_input("Number of nodes: "))
    nodetype = raw_input("Node type [m4.large]: ") or "m4.large"

    use_placement_group = 'n'
    if "nano" not in nodetype and "small" not in nodetype and "micro" not in nodetype: 
        use_placement_group = raw_input("use placement group?: [y]") or 'y'

    spec = {}
    spec['cloud']   = { 'provider' : 'ec2', 'region': cluster.cloud.region }
    spec['cluster'] = { 'name': name }

    if (use_placement_group == 'y' or use_placement_group == 'yes'):
        spec['cluster']['use_placement_group'] = 'yes'

    nodes = []
    def make_node(node_name):
        node = dict()
        node['instance_type'] = nodetype
        node['nodename'] = node_name
        return node

    node = make_node("master")
    nodes.append(node)

    if (numnodes > 1):
        node = make_nodes("worker")
        node['count'] = numnodes-1
        nodes.append(node)

    spec['nodes'] = nodes
    str_yaml = yaml.dump(spec, default_flow_style=False)
    ret = cluster.config.save_cluster_config(name, str_yaml)
    logger.info("Saved cluster spec to: %s" % ret)
    logger.info("Edit this file if needed and run %s$cluster create %s%s" % 
                    (colorama.Fore.GREEN, name, colorama.Style.RESET_ALL))
    cluster.config.read_all_clusters()
    cluster.clusters = cluster.config.get_clusters()

def create_cluster(args, cluster, logger):
    '''
        parse a yaml spec for a cluster and create a cloud formation 
        template using troposphere. create this cluster as a cfn stack.
    '''

    try:
        cluster.config.read_all_clusters()

        cluster_name = args[1]
        clusters = cluster.config.get_clusters()

        obj_yaml = clusters.get(cluster_name)

        if not obj_yaml:
            logger.error("no such cluster spec in %s" % cluster.config.get_clusters_dir())
            return
        
        cloud_spec = obj_yaml.get('cloud')
        target_region = cloud_spec.get('region')

        if target_region == 'closest' or not target_region:
            target_region = get_closest_region(cluster, logger)
            cloud_spec['region'] = target_region

        # for default keys, etc
        if cluster.region != target_region:
            cluster.switch_to_region(target_region)

        # use troposphere to write out a cloud formation template
        cfn_template = Template()
        nodes = obj_yaml.get('nodes')
        nodes = expand_clones(nodes)
        obj_yaml['nodes'] = nodes

        cluster_spec = obj_yaml.get('cluster')
        if not cluster_spec:
            raise Exception("No cluster section in template %s" % specfile)

        cluster_name = cluster_spec.get('name')
        if not cluster_name:
            raise Exception("No cluster name in template %s" % specfile)

        # placement group
        enable_placement = str(cluster_spec.get('use_placement_group')).lower()

        placement_group = None
        if enable_placement == 'yes' or enable_placement == 'true':
            placement_group = ec2.PlacementGroup('DustPlacementGroup')
            placement_group.Strategy='cluster'
            cfn_template.add_resource(placement_group)

            for node in nodes:
                inst_type = node.get('instance_type')
                if "large" not in inst_type:
                    logger.error("Only large instance types can be launched into placement groups.")
                    logger.error("http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/placement-groups.html")
                    return

        # subnet for VPC
        create_vpc = None
        vpc_id     = None
        vpc_subnet = None

        vpc_flag = str(cluster_spec.get('create_vpc')).lower()
        if vpc_flag == 'yes' or vpc_flag == 'true':
            vpc_id, vpc_subnet = configure_vpc(cfn_template, cluster_name)

        elif cluster_spec.get('vpc_id'):
            vpc_id     = cluster_spec.get('vpc_id')
            vpc_subnet = cluster_spec.get('subnet_id')

        if vpc_id and not vpc_subnet:
            logger.error("Need to specify subnet_id and vpc_id if you want to launch nodes into a vpc")
            return

        # create security groups
        node_sec_group, intracluster_sec_group= configure_security_groups(vpc_subnet, vpc_id)
        cfn_template.add_resource(node_sec_group)
        cfn_template.add_resource(intracluster_sec_group)

        # create instances
        have_nano = False 
        for node in nodes:
            nodename = node.get('nodename')
            instance = ec2.Instance(nodename, 
                                    Tags = [ec2.Tag("Name", nodename)])

            keyname = node.get('key')
            if keyname:
                instance.KeyName = keyname
            else:
               # default keys
                default_key = cluster_spec.get('key')
                if not default_key:
                    default_key, keypath = cluster.get_default_key(target_region)
                instance.KeyName = default_key

            image_id = cluster_spec.get('image')
            if image_id:
                instance.ImageId = image_id
                login_user = cluster_spec.get('username')
                if not login_user:
                    logger.error("Need to specify a login user with a custom AMI")
                node['username'] = login_user
            else:
                default_ami, default_login_user = cluster.cloud.get_dust_ami_for_region(target_region)
                instance.ImageId = default_ami
                node['image'] = default_ami
                node['username'] = default_login_user

            instance.InstanceType = node.get('instance_type')

            if 'nano' in instance.InstanceType:
                have_nano = True

            if create_vpc:
                instance.DependsOn = "dustAttachGateway"

            # vpc public ip
            if vpc_subnet:
                network_interface = ec2.NetworkInterfaceProperty("DustNIC" + nodename)
                network_interface.AssociatePublicIpAddress = True
                network_interface.DeleteOnTermination = True
                network_interface.DeviceIndex = 0 
                network_interface.SubnetId = vpc_subnet
                network_interface.GroupSet =  [Ref(node_sec_group)]
                instance.NetworkInterfaces = [network_interface]

            if not vpc_subnet:
                instance.SecurityGroups = [ Ref(node_sec_group) ]

            if placement_group:
                instance.PlacementGroupName = Ref(placement_group)

            cfn_template.add_resource(instance)

        # save it to ./dustcluster/clusters/name_region.cfn
        cfn_json = cfn_template.to_json()

        writecfnto = os.path.join(cluster.config.get_clusters_dir(), "%s.%s.cfn" % (cluster_name, target_region))
        with open(writecfnto, "w") as fh:
            fh.write(cfn_json)

        logger.info("Wrote CloudFormation template to %s" % writecfnto)

        logger.info("Validating template ...")
        conn = get_cfn_connection(logger, cluster, target_region)

        valid = conn.validate_template(cfn_json)
        #print valid.ValidateTemplateResult
        if valid.capabilities:
            print valid.capabilities, valid.capabilities_reason, valid.description

        try:
            if not have_nano: # boto dumps error with nano
                cost_url = conn.estimate_template_cost(cfn_json)
                logger.info("Estimated running cost of this cluster at:  %s" % cost_url)
        except Exception, ex:
            logger.info("Could not estimate template costs.")

        ret = raw_input("Create stack [y]:") or "y"

        if ret.lower()[0] != "y":
            return

        # create the stack

        conn.create_stack(stack_name=cluster_name,  template_body=cfn_json)

        cluster.invalidate_cache()

        save_cluster(cluster, obj_yaml, logger)

    except Exception, e:
        logger.exception('Error: %s' % e)
        logger.error('%sCluster create threw. Please check for new instances with refresh.%s' % (colorama.Fore.RED, colorama.Style.RESET_ALL))
        return

    logger.info( '%sCluster creation kicked off. see status with $cluster status %s.%s' %  
                        (colorama.Fore.GREEN,cluster_name,colorama.Style.RESET_ALL))

    logger.info( 'Refresh node state with the $refresh command.')

def delete_cluster(args, cluster, logger):

    try:

        cluster_name = args[1]

        if cluster_name not in cluster.clusters:
            logger.error("%s is not a cluster" % cluster_name)
            return

        obj_yaml = cluster.clusters[cluster_name]
    
        region = obj_yaml.get('cloud').get('region')

        ret = raw_input("Please confirm: Delete stack[n]? ") or "n"

        if ret.lower()[0] != "y":
            return

        conn = get_cfn_connection(logger, cluster, region)

        # delete the stack
        conn.delete_stack(cluster_name)
        cluster.config.delete_cluster_config(cluster_name, region)
        cluster.config.read_all_clusters()

        cluster.invalidate_cache()

        if cluster.cur_cluster == cluster_name:
            cluster.cur_cluster = ""
        logger.info("Deleted cluster")

    except Exception, e:
        logger.exception('Error: %s' % e)
        return


def expand_clones(nodes):

    ret_nodes = []

    for node in nodes:

        counter = 0
        clones = int(node.get('count') or 1)

        if not node.get('count'):
            ret_nodes.append(node)
        else:
            for i in range(clones):
                newnode = copy.deepcopy(node)
                newnode.pop('count')
                nodename = node.get('nodename')
                nodename_c = "%s%s" % (nodename, i)
                newnode['nodename'] = nodename_c
                ret_nodes.append(newnode)

    return ret_nodes


def save_cluster(cluster, obj_yaml, logger):
    ''' write login rules with new cluster with filters '''

    cluster_props = obj_yaml.get('cluster')
    name = cluster_props.get('name')
    filter_exp = "tags=aws:cloudformation:stack-name:%s" % name

    login_user = cluster_props.get('username') or 'ec2-user'
    region = obj_yaml.get('cloud').get('region')
    args = " %s %s %s h" % (login_user, cluster.get_default_key(region)[1] , name)
    print "ARGs", args
    cluster.handle_command("assign", filter_exp + args)

    cluster.config.read_all_clusters()

def get_cfn_connection(logger, cluster, region):

    cfn_id = 'cloudformation-connection-%s' % region

    conn = cluster.command_state.get(cfn_id)

    if not conn:
        logger.info("Connecting to cloud formation endpoint in %s" % region)
        conn = boto.cloudformation.connect_to_region(region,
                                        aws_access_key_id=cluster.cloud.creds_map['aws_access_key_id'], 
                                        aws_secret_access_key=cluster.cloud.creds_map['aws_secret_access_key'])
        cluster.command_state.put(cfn_id, conn)

    return conn



def cluster_status(args, cluster, logger):
    '''
    cluster status [clustername]  - print notifications from the cloud for this cluster

    Examples:
    cluster status mycluster
    
    Note:
    With no args, describe all stacks _in the current region_
    '''

    try:
        
        cluster_name = None if len(args)<2 else args[1] 

        if not cluster_name:
            conn = get_cfn_connection(logger, cluster, cluster.cloud.region)

            stacks = conn.describe_stacks()
            for stack in stacks:
                print stack
                events = stack.describe_events()
                for event in reversed(events):
                    print event
            return

        obj_yaml = cluster.clusters.get(cluster_name)
        if not obj_yaml:
            logger.error("No such cluster %s" % cluster_name)

        if obj_yaml:
            region = obj_yaml.get('cloud').get('region')
        else:
            region = cluster.cloud.region

        conn = get_cfn_connection(logger, cluster, region)

        # get stack events
        events = conn.describe_stack_events(cluster_name)

        startColorCyan  = "\033[0;36;40m"
        endColor       = "\033[0m"

        for event in reversed(events):
            print event.timestamp, event.resource_type, event.logical_resource_id,\
                event.resource_status, startColorCyan, event.resource_status_reason or "", endColor

    except Exception, e:
        logger.exception('Error: %s' % e)
        return

    logger.info('%sUse $refresh to update node state%s' % (colorama.Fore.GREEN, colorama.Style.RESET_ALL))

    logger.info('ok')



# Copyright (c) Ran Dugal 2014
#
# This file is part of dust.
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

'''
Cluster command engine
'''

import time
import re
import fnmatch
import os
import yaml
import colorama
import sys

from copy import deepcopy

from dustcluster.lineterm import LineTerm
from pkgutil import walk_packages
from dustcluster import commands
from dustcluster.config import DustConfig

from dustcluster.util import setup_logger
logger = setup_logger( __name__ )

import glob

from dustcluster.EC2 import EC2Cloud


class CommandState(object):
    '''
    Commands can store arbitrary state here.  
    '''

    def __init__(self):
        self.global_state   = {}    # { command_name.var : value, where command_name can also be 'global_state' }

    def get(self, state_var):
        return self.global_state.get(state_var)

    def put(self, state_var, value):
        self.global_state[state_var] = value


class ClusterCommandEngine(object):
    '''
    Provides access to the cloud and node objects and methods for resolving a target node name to a list of nodes.
    Loads commands and invokes them. Maintains command state. 
    This object is accessible from all commands. 
    '''

    def __init__(self):

        self.cloud = None
        self.nodecache = dict() # invalidated on load template/start/stop/terminate/etc
        self.region = None

        self.config = DustConfig()

        self.cur_cluster = ""
        self.provider_cache = {}

        self.init_default_provider()

        self.clusters = self.config.get_clusters()

        self._commands = {}
        self.command_state = CommandState()
        self.lineterm = LineTerm()

    def invalidate_cache(self):

        if self.nodecache.get(self.cloud.region):
            del self.nodecache[self.cloud.region]

    def load_commands(self):
        '''
        discover commands under dustcluster.commands relative to this module, and dynamically import command modules
        look for more paths inside ~/.dustcluster/user_commands
        '''
        start_time = time.time()

        cmds = walk_packages( commands.__path__, commands.__name__ + '.')
        cmdmods  = [cmd[1] for cmd in cmds]
        logger.debug( '... loading commands from %s modules' % len(cmdmods))
        logger.debug(list(cmdmods))

        # import commands and add them to the commands map
        for cmdmod in cmdmods:
            try:
                newmod = __import__(cmdmod, fromlist=['commands'])
            except Exception as e:
                logger.error("Error importing command from module [%s] %s" % (cmdmod, e))

            #sys.stdout.write("\rloading .. %-20s          " % cmdmod)
            #sys.stdout.flush()

            for cmdname in newmod.commands:
                cmdfunc =  getattr(newmod, cmdname, 'None')
                if not cmdfunc:
                    logger.error('exported command %s not found in %s' % (cmdname, newmod))
                    continue
                self._commands[cmdname] = (cmdfunc.__doc__, newmod)

        end_time = time.time()

        if self._commands:
            logger.debug('loaded %s commands from %s in %.3f sec\n' % \
                            (len(self._commands), commands.__name__, (end_time-start_time))  )

        for cmd, (shelp, cmdmod) in self._commands.items():
            logger.debug( '%s %s' % (cmdmod.__name__, shelp.split('\n')[1]))

    def get_commands(self):
        return self._commands

    def handle_command(self, cmd, arg):

        try:
            cmddata = self._commands.get(cmd)
            if cmddata:
                _ , cmdmod = cmddata
                func = getattr(cmdmod,  cmd)
                func(arg, self, logger)
                return True
        except Exception, ex:
            logger.exception(ex)
            return True

        return False

    def init_default_provider(self):

        default_region = self.config.get_userdata().get("region")

        cloud_data = { 'provider': 'ec2', 'region' :  default_region}

        self.init_cloud_provider(cloud_data)

    def switch_to_region(self, new_region):

        cloud_data = { 'provider': 'ec2', 'region' : new_region }

        self.unload_cur_cluster()

        if new_region != self.cloud.region:
            self.invalidate_cache()

        self.init_cloud_provider(cloud_data)

        logger.info("Connected to %s " % self.region)


    def switch_to_cluster(self, cluster_name):

        if cluster_name.strip() == '*':
            self.cur_cluster = ""
        else:    
            login_rules = self.config.get_login_rules()
            clusters = set([ rule.get('member-of') for rule in login_rules ])

            if cluster_name not in clusters:
                logger.error("%s is not a recognized cluster." % cluster_name)
                self.show_clusters()
                return

            self.cur_cluster = cluster_name

        cluster_nodes = self.resolve_target_nodes()
        self.show(cluster_nodes)

    def init_cloud_provider(self, cloud_data):

        self.cloud, self.region = self.get_cloud_provider(cloud_data)


    def get_cloud_provider(self, cloud_data):

        provider = cloud_data.get('provider')

        if not provider:
            raise Exception("No 'provider:' section in cloud")

        cloud_provider = None
        cloudregion = cloud_data.get('region')
        if provider.lower() == 'ec2':
            cloud_provider = self.get_cloud_provider_by_region(provider, cloudregion)
        else:
            raise Exception("Unknown cloud provider [%s]." % provider)

        return cloud_provider, cloudregion


    def get_cloud_provider_by_region(self, provider, cloudregion):

        key = (provider,cloudregion)
        cloud_provider = self.provider_cache.get(key) 

        if not cloud_provider:
            cloud_provider = EC2Cloud(creds_map=self.config.get_credentials(), region=cloudregion)
            cloud_provider.connect()
            self.provider_cache[key] = cloud_provider

        return cloud_provider

    def show_clusters(self):
        login_rules = self.config.get_login_rules()
        clusters = set([ rule.get('member-of') for rule in login_rules ])
        logger.info("%sConfigured clusters with logins:%s" % (colorama.Fore.GREEN, colorama.Style.RESET_ALL))
        for cluster in clusters:
            print cluster

    def unload_cur_cluster(self):
        self.cur_cluster = ""

    def get_default_key(self, region=""):
        ''' get the default key for current or specified region

            first we go to user_data and get the default key name for this install
            if its not there, create the name
                              creaet the defauly key name entry
                              create the key
                                 -- if it already exists, show a warning
                              create the mapping in user_data

            query user data for the keys location
        '''

        try:

            if not region:
                region = self.cloud.region

            user_data = self.config.get_userdata()
            default_keynames = user_data.get('default-keynames') or {}

            namekey = 'ec2-default-keyname-' + region 
            default_keyname = ""
            if default_keynames:
                default_keyname = default_keynames.get(namekey) or {}

            if not default_keynames or not default_keyname:
                timestamp = time.strftime("%Y%m%d%H%M%S", time.gmtime())
                default_keyname = "dust_%s_%s" % (region, timestamp)
                default_keyname = default_keyname.replace("-", "")
                default_keynames[namekey] = default_keyname

                exists, keyname, keypath = self.cloud.create_keypair(default_keyname, self.config.get_default_keys_dir())

                self.config.get_userdata().update( { 'default-keynames' :  default_keynames } )
                self.config.write_user_data()

                region_key = "%s#%s" % (region, default_keyname)

                if exists and not keypath:
                    errstr = "WARNING: The default key %s exists in the cloud but wasnt created with the local dustcluster install. " % keyname
                    errstr += "You will be asked for the key file while attempting ssh operations into this cluster. "
                    errstr += "If you do not have access to this key, specify another key in the cluster config. "
                    logger.warn("%s%s%s" % (colorama.Fore.RED, errstr, colorama.Style.RESET_ALL))

                if keypath:
                    keymap = user_data.get('ec2-key-mapping') or {}
                    keymap[region_key] = keypath
                    userdata = self.config.get_userdata()
                    userdata.update( { 'ec2-key-mapping':  keymap } )
                    logger.info("Updating new key mappings to userdata [%s]" % self.config.userdata_file)
                    self.config.write_user_data()

                return default_keyname, keypath

            keymap = self.config.get_userdata().get('ec2-key-mapping') or {}
            region_key = "%s#%s" % (region, default_keyname)
            keyfile = keymap.get(region_key)
            if not keyfile:
                errstr = "default key %s exists in user_data mapping but the file cannot be found" % default_keyname
                logger.error("%s%s%s" % (colorama.Fore.RED, errstr, colorama.Style.RESET_ALL))

            return default_keyname, keyfile

        except Exception, ex:
            logger.exception(ex)
            logger.error('Error getting default keys: %s' % ex)


    def show(self, nodes, extended=False):
        ''' print summary of cloud nodes to stdout '''

        if not self.cloud:
            logger.error('Internal error. No cloud provider set.')
            return

        # get node data
        if not nodes:
            logger.debug("No nodes to show")
            return []

        logger.info("Nodes in region: %s" % self.cloud.region)

        try:
            header_data, header_fmt = nodes[0].disp_headers()

            print colorama.Fore.GREEN
            print "   ", " ".join(header_fmt) % tuple(header_data)
            print colorama.Style.RESET_ALL

            prev_cluster_name = "_"
            prev_vpc = "_"
            for node in nodes:
                
                node_vpc = node.get('vpc')
                if node_vpc and node_vpc != prev_vpc:
                    print ""
                    print "%s--%s:%s" % (colorama.Fore.GREEN, node_vpc, colorama.Style.RESET_ALL)
                    prev_vpc = node_vpc

                if node.cluster != prev_cluster_name:
                    if node.cluster:
                        if True:
                            print( "%scluster [%s]" % (colorama.Style.RESET_ALL, node.cluster))
                        else:
                            print( "%s%s" % (colorama.Style.RESET_ALL, node.cluster))
                        prev_cluster_name = node.cluster or cluster_filter
                    else:
                        print colorama.Style.RESET_ALL
                        print( "unassigned:" )
                        prev_cluster_name = None

                print colorama.Style.NORMAL, colorama.Fore.CYAN, " ", " ".join(header_fmt) % tuple(node.disp_data())
                ext_data = []
                if extended == 1:
                    ext_data = node.extended_data().items()
                elif extended == 2:
                    ext_data = node.all_data().items()

                if ext_data:
                    for (k,v) in sorted(ext_data, key=lambda x:x[0]):
                        print colorama.Style.RESET_ALL, colorama.Style.DIM, header_fmt[1] % "", k, ":", v
                    print colorama.Style.RESET_ALL

        finally:
            print colorama.Style.RESET_ALL


    def _filter_tags(self, tags, fkey, fval):

        keyregex = fnmatch.translate(fkey)
        keymatch = re.compile(keyregex)

        valregex = fnmatch.translate(fval)
        valmatch = re.compile(valregex)

        # match tag keys, then vals
        for tagkey, tagval in tags.items():
            if keymatch.match(tagkey.lower()) and valmatch.match(tagval.lower()):
                return True

        return False


    def _filter(self, nodes, filterkey, filterval):
        '''
        filter a list of nodes by attribute values
        e.g. filterkey=state,  filterval=running
        '''

        filtered = []

        if not filterkey:
            return nodes

        filterkey = filterkey.lower()
        filterval = filterval.lower()

        if filterkey == "tags":

            fkey = ""
            fval = ""

            pos = filterval.rfind(':')
            if pos:
                fval = filterval[pos+1:]
                fkey = filterval[:pos]

                if fkey[0] == '"' and fkey[-1] == '"':
                    fkey = fkey[1:-1]

            if not fkey and not fval:
                logger.error("Bad filter. Use tags=key:value, wildcards allowed on key, value.")
                return []

            for node in nodes:
                if self._filter_tags(node.get('tags'), fkey, fval):
                    filtered.append(node)

        else:

            regex = fnmatch.translate(filterval.lower())
            valid = re.compile(regex)

            for node in nodes:
                val =  node.get(filterkey)
                if not val:
                    continue
                logger.debug("trying filter %s on %s..." % (filterval, val))
                if not valid.match(str(val).lower()):
                    continue
                filtered.append(node)

        return filtered

    def _search(self, nodes, search_term):
        '''
        filter a list of nodes by search term
        e.g. "running"
        '''

        results = set()
        search_term = search_term.lower()

        for node in nodes:
            for val in node.get('tags').values():
                if search_term in val.lower():
                    results.add(node)
                    break

            for key in node.extended_fields + node.all_fields:
                if search_term in str(node.get(key)).lower():
                    results.add(node)
                    break

        return list(results)

    def _find_matching_nodes(self, selector, cluster_nodes):
        ''' using node_props.selector, find the matching node in cluster_nodes ''' 

        if selector == '*':
            return cluster_nodes

        if '=' not in selector:
            logger.error("Invalid login selector %s" % selector)
            return []

        filterkey, filterval = selector.split("=")

        #logger.debug("matching template node filters [%s=%s]to cluster nodes" % (filterkey, filterval))
        matching_nodes = self._filter(cluster_nodes, filterkey, filterval)

        return matching_nodes

    def _get_nodes_with_login_data(self):
        ''' get all nodes matched to login rules '''

        nodecache_nodes = self.nodecache.get(self.cloud.region)
        if nodecache_nodes:
            logger.info("Retrieved [%d] nodes %s%sfrom cache%s" % (len(nodecache_nodes),
                                                                    colorama.Fore.GREEN, colorama.Style.BRIGHT, colorama.Style.RESET_ALL))
            nodes = nodecache_nodes
        else:
            nodes = self.cloud.refresh()
            logger.info("Retrieved [%d] nodes %s%sfrom cloud provider%s" % (len(nodes), 
                                                                    colorama.Fore.GREEN, colorama.Style.BRIGHT, colorama.Style.RESET_ALL))
            self.nodecache[self.cloud.region] = nodes

        self._match_nodes_to_login_rules(nodes)
        ret_nodes = sorted(nodes, key =lambda x: (x.cluster, x.get('vpc')))
        for i in range( len(ret_nodes) ):
            ret_nodes[i].index = str(i + 1)
        return ret_nodes

    def _match_nodes_to_login_rules(self, nodes):

        for rule in self.config.get_login_rules():
            selector = rule.get('selector')
            matched_nodes = self._find_matching_nodes(selector, nodes)
            for node in matched_nodes:
                if not node.login_rule:         # to maintain rule precedence, dont rematch
                    node.login_rule = rule
                    node.cluster = rule.get('member-of')

    def resolve_target_nodes(self, search=False, target_node_name=""):
        '''
            given a target string, filter nodes from cloud provider and create a list of target nodes to operate on
            filter by all clusters
        '''
        if not self.cloud:
            raise Exception('Internal error: No cloud provider loaded.')

        cluster_nodes = self._get_nodes_with_login_data()

        # working set
        if self.cur_cluster:
            cluster_nodes = filter(lambda x: (x.login_rule.get('member-of') == self.cur_cluster), cluster_nodes)

        if not cluster_nodes:
            return []

        # filter by target string
        # target string can be a comma separated list of filter expressions

        if target_node_name == '*' or not target_node_name:
            return cluster_nodes

        filtered_nodes = set()
        for filter_exp in target_node_name.split(","):
            if '=' in filter_exp:
                parts = filter_exp.split("=")
                if len(parts) != 2:
                    str_err = "Filter format error. Should be key=value for node properties or tags=key:value for tags. Use quotes if needed."
                    raise Exception("%s%s%s" %(colorama.Fore.RED, str_err, colorama.Style.RESET_ALL))
                filterkey, filterval  = parts
            elif filter_exp in self.clusters:
                filterkey, filterval = 'cluster', filter_exp    
            else:
                filterkey, filterval = 'name', filter_exp
                if self.is_numeric(filter_exp):
                    filterkey = 'index'

            if filterkey and filterval:
                target_nodes  = self._filter(cluster_nodes, filterkey, filterval)
                filtered_nodes.update(target_nodes)

        if search and not filtered_nodes:
            for search_term in target_node_name.split(","):
                filtered_nodes.update( self._search(cluster_nodes, search_term))

        if not filtered_nodes:
            logger.info( "no nodes found that match filter %s" % (target_node_name) )
        else:
            logger.debug( "resolved target string [%s] to %s nodes" % (target_node_name, len(filtered_nodes)) )

        return sorted( list(filtered_nodes), key=lambda x: x.index )


    def running_nodes_from_target(self, target_str):
        '''
        params: same as a command
        returns: target_nodes : a list of running target nodes
        '''

        target_nodes = self.any_nodes_from_target(target_str)
        if not target_nodes:
            return None

        target_nodes = [node for node in target_nodes if node.get('state') == 'running']
        if not target_nodes:
            logger.info( 'No target nodes are in the running state' )
            return None

        return target_nodes


    def any_nodes_from_target(self, target_str):
        '''
        params: same as a command
        returns: target_nodes : a list of stopped or running target nodes
        '''

        target_nodes = self.resolve_target_nodes(target_node_name=target_str)
        if not target_nodes:
            logger.info( 'No target nodes found that match filter [%s]' % target_str)

        return target_nodes

    def logout(self):
        self.lineterm.shutdown()

    def is_numeric(self, s):
        try:
            int(s)
            return True
        except ValueError:
            return False



# Copyright (c) Ran Dugal 2014
#
# This file is part of dust.
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
# 
# This program is distributed in the hope that it will be useful, but WITHOUT 
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

''' dust command for getting and putting files from/to a set of nodes '''

import glob

# export commands

commands = ['put', 'get']

def put(cmdline, cluster, logger):
    '''
    put localfiles filter [target dir] - upload local files

    Notes:
    localfiles can have wildcards

    Examples:
    put /opt/data/data1.txt worker* # uploads data.txt to cwd
    put /opt/data/data*.txt 1,2     # wildcards work
    put /opt/data/data2.txt worker* /opt/data
    '''
    if not cmdline or len(cmdline) < 2:
        logger.error("usage: put src filter [dest]")
        return

    arr  = cmdline.split()
    srcfile = arr[0]
    target = arr[1]

    destdir = None
    if len(arr) > 2:
        destdir = arr[2]

    globbed = glob.iglob(srcfile)
    if not list(globbed):
        logger.error('local files not found : %s' % srcfile)
        return

    target_nodes = cluster.running_nodes_from_target(target)
    if not target_nodes:
        return

    for node in target_nodes:
        globbed = glob.iglob(srcfile)
        for fname in globbed: 
            cluster.lineterm.put(node.login_rule.get('keyfile'), node, fname, destdir)


def get(cmdline, cluster, logger):
    '''
    get filter remotefile [localdir] - download remotefile from tgt nodes as remotefile.nodename

    Notes:
    remotefile cannot be a wildcard

    Example:
    get worker* /opt/output/data1.txt    # download to cwd
    get worker* /opt/output/data1.txt /tmp   # download to /tmp
    '''

    if not cmdline or len(cmdline) < 2:
        logger.error("usage: get filter remotefile [localdir]")
        return

    target = cmdline.split()[0]

    target_nodes = cluster.running_nodes_from_target(target)
    if not target_nodes:
        logger.info('no running nodes match %s' % target)
        return

    args = cmdline[len(target):].strip()

    arrargs = args.split()

    remotefile = None
    localdir = None

    remotefile = arrargs[0]

    if len(arrargs) > 1:
        localdir = arrargs[1]

    for node in target_nodes:
        cluster.lineterm.get(cluster.cloud.keyfile, node, remotefile, localdir)


